import functools
import io
import re
import socket
import tarfile
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from ownca import CertificateAuthority
from typing import Callable
import asyncio
import time
from collections import deque
import statistics
import math
import itertools
from rich.pretty import pretty_repr
from copy import copy

from rich.align import Align

import dns.resolver
import dns.reversename
import keyring
import randomname
from fluxrpc.client import RPCClient
from fluxrpc.auth import SignatureAuthProvider
from fluxrpc.transports.socket.symbols import (
    AUTH_ADDRESS_REQUIRED,
    AUTH_DENIED,
    NO_SOCKET,
    PROXY_AUTH_ADDRESS_REQUIRED,
    PROXY_AUTH_DENIED,
)

from fluxvault.log import log
from rich.pretty import pprint


async def handle_session(agent):
    if agent.transport.session.started:
        if not agent.transport.session.connection_attempted:
            await agent.transport.session.start(connect=True)

            if signing_address := agent.transport.session.signing_address:
                signing_key = keyring.get_password("fluxvault_app", signing_address)

                if not signing_key:
                    log.error(f"Signing key required in keyring for {signing_address}")
                    raise FluxVaultKeyError(
                        f"Reason: {agent.transport.failed_on} Signing key for address: {getattr(agent.transport, address)} not present in secure storage"
                    )

                await agent.transport.session.connect(signing_key)


async def handle_connection(agent: RPCClient, connect: bool, exclusive: bool = False):
    if connect:
        await agent.transport.connect(
            exclusive=exclusive
        )  # this gives us exclusive use of channel
        if agent.transport.connected:
            return

    if not agent.transport.connected:
        log.info("Transport not connected... checking connection requirements...")
        log.info(f"Failed on {agent.transport.failed_on}")

        if agent.transport.failed_on == NO_SOCKET:
            return

        address = ""
        # match/case
        if agent.transport.failed_on in [AUTH_ADDRESS_REQUIRED, AUTH_DENIED]:
            address = "auth_address"
        elif agent.transport.failed_on in [
            PROXY_AUTH_ADDRESS_REQUIRED,
            PROXY_AUTH_DENIED,
        ]:
            address = "proxy_auth_address"

        signing_key = keyring.get_password(
            "fluxvault_app", getattr(agent.transport, address)
        )

        if not signing_key:
            log.error(
                f"Signing key required in keyring for {getattr(agent.transport, address)}"
            )
            raise FluxVaultKeyError(
                f"Reason: {agent.transport.failed_on} Signing key for address: {getattr(agent.transport, address)} not present in secure storage"
            )

        auth_provider = SignatureAuthProvider(key=signing_key)
        agent.transport.auth_provider = auth_provider
        await agent.transport.connect(exclusive=True)


# def manage_session(f):
#     @functools.wraps(f)
#     async def wrapper(*args, **kwargs):
#         # Surely there is a better way
#         agent = None
#         for arg in args:
#             if isinstance(arg, RPCClient):
#                 agent = arg
#                 break

#         if not agent:
#             return

#         if agent.transport.session.started:
#             if not agent.transport.session.connection_attempted:
#                 await agent.transport.session.start(connect=True)

#                 if signing_address := agent.transport.session.signing_address:
#                     signing_key = keyring.get_password("fluxvault_app", signing_address)

#                     if not signing_key:
#                         log.error(
#                             f"Signing key required in keyring for {signing_address}"
#                         )
#                         raise FluxVaultKeyError(
#                             f"Reason: {agent.transport.failed_on} Signing key for address: {getattr(agent.transport, address)} not present in secure storage"
#                         )

#                     await agent.transport.session.connect(signing_key)

#         # print("wrapper args", args)
#         # print("wrapper kwargs", kwargs)

#         # try:
#         res = await f(*args, **kwargs)
#         # except asyncio.TimeoutError:
#         #     # Not sure about this
#         #     agent.transport.failed_on = NO_SOCKET
#         #     agent.transport.connected = False
#         #     res = None
#         #     log.error(f"Timeout error waiting for response from: {f.__name__}")

#         return res

#     return wrapper


def manage_transport(f=None, exclusive: bool = False):
    def inner(f):
        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            # Surely there is a better way
            agent = None
            for arg in args:
                if isinstance(arg, RPCClient):
                    agent = arg
                    break

            disconnect = kwargs.pop("disconnect", True)
            connect = kwargs.pop("connect", True)
            in_session = kwargs.pop("in_session", False)

            print(
                f"In wrapper for {f.__name__}, connect: {connect}, disconnect: {disconnect}, in_session: {in_session} agent: {agent.id}"
            )
            try:
                # they shoudl probably both have a channel id
                if in_session:
                    await handle_session(agent)
                else:
                    await handle_connection(agent, connect, exclusive)

                if not agent.transport.connected:
                    log.error("Connection failed... returning")
                    return

                res = await f(*args, **kwargs)

                if not in_session and disconnect:
                    await agent.transport.disconnect()

                return res
            except Exception as e:
                print(repr(e))
                exit(0)

        return wrapper

    if f:
        return inner(f)
    else:
        return inner


class SyncStrategy(Enum):
    STRICT = 1
    ALLOW_ADDS = 2
    ENSURE_CREATED = 3


class AppMode(Enum):
    FILESERVER = 1
    SINGLE_COMPONENT = 2
    MULTI_COMPONENT = 3
    UNKNOWN = 4


class ContainerState(Enum):
    DEFAULT = "DEFAULT"
    ERROR = "ERROR"
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    UNCONTACTABLE = "UNCONTACTABLE"


def bytes_to_human(num, suffix="B"):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def human_to_bytes(size: str):
    # macOS etc
    units = {"B": 1, "KB": 10**3, "MB": 10**6, "GB": 10**9, "TB": 10**12}

    # Alternative unit definitions, notably used by Windows:
    # units = {"B": 1, "KB": 2**10, "MB": 2**20, "GB": 2**30, "TB": 2**40}

    number, unit = list(filter(None, re.split("(\d+)", size)))
    return int(float(number) * units[unit])


def tar_object(dir_or_file: Path) -> bytes:
    log.info(f"About to tar {dir_or_file}")
    fh = io.BytesIO()
    # lol, hope the files aren't too big, or, you got plenty of ram
    with tarfile.open(fileobj=fh, mode="w|bz2") as tar:
        tar.add(
            dir_or_file,
            arcname="",
        )
    log.info("Tarring complete")
    return fh.getvalue()


def size_of_object(path: Path) -> int:
    obj_type = "File"
    # ToDo: this breaks if path doesn't exist
    if path.is_dir():
        obj_type = "Directory"

        size = sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())

    elif path.is_file():
        size = path.stat().st_size

    # this feels wrong here
    log.info(f"Syncing {obj_type} {path} of size { bytes_to_human(size)}")

    return size


class AgentId(tuple):
    ...


class UnixTime(float):
    ...


class SymbolDeque(deque):
    def __init__(self, symbols: list = ["\U0001F7E9", "\U0001F7E5"]):
        self.update_symbols(symbols)

    def update_symbols(self, symbols: list):
        # meh, IndexOutOfRange
        self.positive = symbols[0]
        self.negative = symbols[1]

    def append(self, item):
        if item:
            super(SymbolDeque, self).append(self.positive)
        else:
            super(SymbolDeque, self).append(self.negative)


@dataclass
class HitCounter:
    raw: deque[bool] = field(default_factory=lambda: deque([], maxlen=60))
    one_minute: deque[bool] = field(default_factory=lambda: deque([], maxlen=15))
    fifteen_minute: deque[bool] = field(default_factory=lambda: deque([], maxlen=15))
    one_hour: deque[bool] = field(default_factory=lambda: deque([], maxlen=15))
    last_update: UnixTime = 0
    counter: int = 0

    # def __repr__(self):
    #     me = self.__dict__.copy()
    #     del me["raw"]
    #     return pretty_repr(me)

    def update(self, hit: bool):
        changed = False
        self.counter += 1
        raw_copy = copy(self.raw)
        self.raw.append(hit)

        if raw_copy != self.raw:
            changed = True

        if self.counter % (60 * 1) == 0:
            one_min_copy = copy(self.one_minute)
            self.one_minute.append(all(self.raw))
            if one_min_copy != self.one_minute:
                changed = True

        if self.counter % (60 * 15) == 0:
            fifteen_min_copy = copy(self.fifteen_minute)
            self.fifteen_minute.append(all(self.one_minute))
            if fifteen_min_copy != self.fifteen_minute:
                changed = True

        if self.counter % (60 * 60) == 0:
            one_hour_copy = copy(self.fifteen_minute)
            self.one_hour.append(all(self.fifteen_minute))
            if one_hour_copy != self.one_hour:
                changed = True

            self.counter = 0

        if changed:
            self.last_update = time.monotonic()


@dataclass
class RTT(tuple):
    raw: deque[float] = field(default_factory=lambda: deque([], maxlen=60))
    one_minute: deque[float] = field(default_factory=lambda: deque([], maxlen=15))
    fifteen_minute: deque[float] = field(default_factory=lambda: deque([], maxlen=4))
    one_hour: deque[float] = field(default_factory=lambda: deque([], maxlen=24))
    low: float = 0
    high: float = 0
    average: float = 0

    def __repr__(self):
        me = self.__dict__.copy()
        del me["raw"]
        return pretty_repr(me)


@dataclass
class StateTransition:
    offline_to_online: bool
    time: UnixTime = field(default_factory=time.time)


@dataclass
class NodeContactState:
    first_contact: UnixTime = field(default_factory=time.time)
    transitions: list[StateTransition] = field(default_factory=list)
    in_quarantine: bool = False
    quarantine_timer: UnixTime = 0
    total_count: int = 0
    total_missed_count: int = 0
    rtt: RTT = field(default_factory=RTT)
    hit_counter: HitCounter = field(default_factory=HitCounter)
    misses: deque[float] = field(default_factory=lambda: deque([], maxlen=60))

    @property
    def active(self) -> bool:
        if self.transitions:
            return self.transitions[-1].offline_to_online
        else:
            return True

    @property
    def transition_count(self) -> int:
        return len(self.transitions)

    @property
    def latest_transition(self) -> StateTransition | None:
        if self.transitions:
            return self.transitions[-1]

    @property
    def last_state_change_time(self) -> UnixTime:
        return 0 if not self.transitions else self.latest_transition.time

    @property
    def one_minute_miss_count(self) -> int:
        count = 0
        one_minute_ago = time.time() - 60

        for miss in reversed(self.misses):
            if miss > one_minute_ago:
                count += 1
            else:
                break

        return count

    # def hit_counter(self) -> list[bool]:
    #     interval = time.time() - 60
    #     heartbeat = [True] * 60
    #     for miss in reversed(self.misses):
    #         if miss > interval:
    #             in_second = miss - interval
    #             in_second = math.floor(in_second)
    #             heartbeat[in_second] = False
    #         else:
    #             break

    #     return heartbeat

    def increment_counters(self):
        self.total_missed_count += 1
        self.misses.append(time.time())

    def quarantine(self):
        self.in_quarantine = True
        self.quarantine_timer = time.perf_counter()

    def dequarantine(self):
        self.in_quarantine = False
        self.quarantine_timer = 0

    def update_hit_counter(self, hit: bool):
        self.hit_counter.update(hit)

    def update_rtt(self, rtt: float):
        rtt = round(rtt, 3)
        self.rtt.raw.append(rtt)

        if self.total_count == 0:
            return

        if self.total_count % (60 * 1) == 0:
            one_min_average = statistics.mean(self.rtt.raw)
            self.rtt.one_minute.append(round(one_min_average, 3))

        if self.total_count % (60 * 15) == 0:
            fifteen_min_average = statistics.mean(self.rtt.one_minute)
            self.rtt.fifteen_minute.append(round(fifteen_min_average, 3))

        if self.total_count % (60 * 60) == 0:
            one_hour_average = statistics.mean(self.rtt.fifteen_minute)
            self.rtt.one_hour.append(round(one_hour_average, 3))

        rtt_count = self.total_count - self.total_missed_count

        # average block
        if self.rtt.average == 0:
            self.rtt.average = rtt
        else:
            rtt_total = rtt_count * self.rtt.average
            rtt_total + rtt
            self.rtt_average = round(rtt_total / rtt, 3)

        # high / low block
        if self.rtt.low == 0 or rtt < self.rtt.low:
            self.rtt.low = rtt

        elif rtt > self.rtt.high:
            self.rtt.high = rtt


@dataclass
class FluxTask:
    name: str
    args: tuple = ()
    kwargs: dict = field(default_factory=dict)
    func: Callable | None = None


@dataclass
class FluxVaultContext:
    agents: dict
    ca: CertificateAuthority
    storage: dict = field(default_factory=dict)


@dataclass
class RemoteStateDirective:
    """"""

    name: str | None = None
    content_source: Path | None = None
    remote_dir: Path | None = None
    sync_strategy: SyncStrategy = SyncStrategy.ENSURE_CREATED

    # @property
    # def absolute_dir(self):
    #     if not any([self.workdir, self.prefix]):
    #         ...
    #     # maybe do some sanity stuff here, make sure at least one of them are absolute
    #     match self.prefix:
    #         case x if x and x.is_absolute():
    #             return x
    #         case x if x:
    #             return self.workdir / self.prefix
    #         case x if not x:
    #             return self.workdir

    @property
    def local_absolute_path(self):
        return None if not self.content_source else self.content_source / self.name

    def serialize(self):
        return {
            "content_source": str(self.content_source),
            "prefix": str(self.remote_dir),
            "sync_strategy": self.sync_strategy.name,
        }


class FluxVaultKeyError(Exception):
    pass


def _get_own_ip() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't have to be reachable
        s.connect(("10.254.254.254", 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = "127.0.0.1"
    finally:
        s.close()
    return ip


def _get_ptr(ip: str) -> str:
    canonical = dns.reversename.from_address(ip)
    resolver = dns.resolver.Resolver()
    try:
        answer = resolver.resolve(canonical, "PTR")
    except dns.resolver.NXDOMAIN:
        return ""
    else:
        return answer[0].to_text()


def _parse_ptr_to_names(ptr: str) -> list:
    # The ptr record contains the fqdn - hostname.networkname
    if not ptr or ptr == "localhost.":
        return ["", ""]

    app_name = ""
    fqdn = ptr.split(".")
    fqdn = list(filter(None, fqdn))
    host = fqdn[0]
    host = host.lstrip("flux")
    host = host.split("_")
    component_name = host[0]
    # if container name isn't specified end up with ['15f4fcb5a668', 'http']
    if len(host) > 1:
        app_name = host[1]
    return [component_name, app_name]


def get_app_and_component_name(_ip: str | None = None) -> list:
    """Gets the component and app name for a given ip. If no ip is given, gets our own details"""
    ip = _ip if _ip else _get_own_ip()
    ptr = _get_ptr(ip)
    comp, app = _parse_ptr_to_names(ptr)

    if not comp or not app:
        comp = randomname.get_name()
        app = "testapp"

    return comp, app



# Standard library

# So you don't have to forward reference in typing own class. See FsEntry
from __future__ import annotations

import asyncio
import functools
import shutil
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Callable
import time
from contextlib import asynccontextmanager

import itertools
import aiofiles

# 3rd party
import aiohttp
import cryptography
import yaml
from cryptography.x509.oid import NameOID

# from rich.pretty import pretty_repr
# from pprint import pformat
from fluxrpc.auth import SignatureAuthProvider
from fluxrpc.client import RPCClient, RPCProxy
from fluxrpc.exc import MethodNotFoundError
from fluxrpc.protocols.jsonrpc import JSONRPCProtocol
from fluxrpc.transports.socket.client import EncryptedSocketClientTransport
from fluxrpc.transports.socket.symbols import NO_SOCKET, PROXY_NO_SOCKET
from ownca import CertificateAuthority
from ownca.exceptions import OwnCAInvalidCertificate
from rich.pretty import pprint, pretty_repr
from rich.align import Align

# this package
from fluxvault.app_init import setup_filesystem_and_wallet
from fluxvault.constants import WWW_ROOT
from fluxvault.fluxapp import (
    FluxApp,
    FluxComponent,
    # FluxTask,
    FsEntryStateManager,
    FsStateManager,
    RemoteStateDirective,
)
from fluxvault.fluxkeeper_gui import FluxKeeperGui
from fluxvault.helpers import (
    AppMode,
    NodeContactState,
    StateTransition,
    FluxVaultKeyError,
    SyncStrategy,
    FluxVaultContext,
    FluxTask,
    AgentId,
    HitCounter,
    UnixTime,
    bytes_to_human,
    manage_transport,
)
from fluxvault.log import log

CONFIG_NAME = "config.yaml"

# path types
#                  absolute  | relative | relative      | absolute
# full_fake_root = vault_dir / app_dir  / fake_root_dir / remote_dir
# app_dir is portable
# The only common format is absolute_remote
# only way to convert back and forward is with managed_object


class FluxKeeper:
    """Runs in your protected environment. Provides runtime
    data to your vulnerable services in a secure manner

    The end goal is to be able to secure an application's private data where visibility
    of that data is restricted to the application owner
    """

    # GUI hidden via cli, no where near ready, look at just breaking out console first
    def __init__(
        self,
        vault_dir: str | None = None,
        apps: FluxApp | None = None,
        # gui: bool = False,
    ):
        # ToDo: configurable port
        self.gui = FluxKeeperGui("127.0.0.1", 7777, self)

        self.loop = asyncio.get_event_loop()
        self.managed_apps: list[FluxAppManager] = []
        self.root_dir: Path = setup_filesystem_and_wallet()

        self.qualify_vault_dir(vault_dir)
        self.apps: list[FluxApp] = []

        # Allow apps to be passed in, otherwise - look up config
        for app in apps:
            if isinstance(app, FluxApp):
                self.apps.append(app)

        if not self.apps:
            for app_dir in self.vault_dir.iterdir():
                if not app_dir.is_dir():
                    continue

                try:
                    with open(app_dir / CONFIG_NAME, "r") as stream:
                        try:
                            config = yaml.safe_load(stream)
                        except yaml.YAMLError as e:
                            raise ValueError(
                                f"Error parsing vault config file: {CONFIG_NAME} for app {app_dir}. Exc: {e}"
                            )
                except (FileNotFoundError, PermissionError) as e:
                    log.error(
                        f"Error opening config file {CONFIG_NAME} for app {app_dir}. Exc: {e}"
                    )
                    continue

                self.apps.append(
                    self.build_app(app_dir.name, self.vault_dir / app_dir.name, config)
                )

        log.info(f"App Data directory: {self.root_dir}")
        log.info(f"Global Vault directory: {self.vault_dir}")
        log.info(f"Apps loaded: {[x.name for x in self.apps]}")

        self.init_certificate_authority()
        self.configure_apps()

        # if gui:
        #     self.start_gui()

    @classmethod
    def setup(cls) -> Path:
        return setup_filesystem_and_wallet()

    def qualify_vault_dir(self, dir: str):
        """Sets the vault_dir attr"""
        if not dir:
            vault_dir = Path().home() / ".vault"
        else:
            vault_dir = Path(dir)

        if not vault_dir.is_absolute():
            raise ValueError(f"Invalid vault dir: {vault_dir}, must be absolute")

        if not vault_dir.is_dir():
            vault_dir.mkdir(parents=True)

        self.vault_dir = vault_dir

    @classmethod
    def state_directives_builder(
        cls, local_relative: Path, remote_workdir: Path, fs_entries: list
    ) -> list:
        state_directives = []
        for fs_entry in fs_entries:
            parent = None
            name = fs_entry.get("name", None)

            if content_source := fs_entry.get("content_source", None):
                content_source = Path(content_source)
                name = content_source.name
                parent = content_source.parent
            else:
                # this is debatable, maybe simplier just to make it if you
                # want to manipulate a files location in the tree, you must
                # supply the content source.

                # try the root of the staging dir
                parent = local_relative

            if remote_dir := fs_entry.get("remote_dir"):
                if Path(remote_dir).is_absolute():
                    absolute_dir = Path(remote_dir)
                else:
                    absolute_dir = remote_workdir / remote_dir
            else:
                absolute_dir = remote_workdir

            sync_strategy = SyncStrategy[
                fs_entry.get("sync_strategy", SyncStrategy.ENSURE_CREATED.name)
            ]

            state_directive = RemoteStateDirective(
                name, parent, absolute_dir, sync_strategy
            )

            state_directives.append(state_directive)
        return state_directives

    # do this as a lambda?
    # flux_tasks = []
    # map(lambda x: flux_tasks.append(FluxTask(x.get("name"), x.get("params"))), tasks)
    # @classmethod
    # def tasks_builder(cls, tasks: list) -> list:
    #     flux_tasks = []
    #     for task in tasks:
    #         flux_task = FluxTask(task.get("name"), task.get("params"))
    #         flux_tasks.append(flux_task)
    #     return flux_tasks

    @classmethod
    def build_app(cls, name: str, app_dir: str, config: dict) -> FluxApp:
        log.info(f"User config:\n")
        pprint(config)
        print()

        app_config = config.get("app_config")
        groups_config = app_config.pop("groups", None)

        app = FluxApp(name, root_dir=app_dir, **app_config)

        components_config = config.get("components", {})

        for component_name, directives in components_config.items():
            remote_workdir = directives.pop("remote_workdir")

            if not Path(remote_workdir).is_absolute():
                raise ValueError(f"Remote workdir {remote_workdir} is not absolute")

            # if app.app_mode == AppMode.FILESERVER:
            #     local_workdir = app_dir
            # else:
            local_workdir = app_dir / "components" / component_name

            # this is if the content source doesn't exist... we try here
            relative_dir = Path("components") / component_name / "staging"

            component = FluxComponent(
                component_name,
                local_workdir=local_workdir,
                remote_workdir=Path(remote_workdir),
            )

            # add the members to the component, then fix up build_catalogue

            # ummmmm lol
            if groups := directives.pop("member_of", None):
                component.add_groups(groups)
                for group in groups:
                    if g := groups_config.get(group, None):
                        if d := g.get("state_directives", None):
                            if directives.get("state_directives", None):
                                directives["state_directives"].extend(d)
                            else:
                                directives["state_directives"] = d

            for directive, data in directives.items():
                match directive:
                    case "state_directives":
                        component.state_manager.add_directives(
                            FluxKeeper.state_directives_builder(
                                relative_dir, Path(remote_workdir), data
                            )
                        )
                    case "tasks":
                        component.add_tasks(FluxKeeper.tasks_builder(data))
            app.add_component(component)

        log.info("Built app config:\n")
        pprint(app)
        print()
        return app

    def init_certificate_authority(self):
        common_name = "keeper.fluxvault.com"

        self.ca = CertificateAuthority(
            ca_storage=f"{str(self.root_dir / 'ca')}", common_name="Fluxvault Keeper CA"
        )

        try:
            cert = self.ca.load_certificate(common_name)
        except OwnCAInvalidCertificate:
            cert = self.ca.issue_certificate(common_name, dns_names=[common_name])

        self.cert = cert.cert_bytes
        self.key = cert.key_bytes
        self.ca_cert = self.ca.cert_bytes

    def configure_apps(self):
        for app in self.apps:
            # if app.app_mode == AppMode.FILESERVER:
            #     app.build_fs()
            # else:
            app.build_catalogue()
            app.validate_local_objects()
            flux_app = FluxAppManager(self, app)
            self.managed_apps.append(flux_app)

    def start_gui(self):
        self.loop.run_until_complete(self.gui.start())

    def cleanup(self):
        log.info("Fluxkeeper cleanup called...")
        for app in self.apps:
            app.remove_catalogue()

        for app in self.managed_apps:
            app.cleanup()

    def get_app_manager_by_name(self, name: str) -> FluxAppManager:
        return next(filter(lambda x: x.app.name == name, self.managed_apps), None)

    async def manage_apps(self, run_once: bool, polling_interval: int):
        tasks = []

        async def manage(app_manager: FluxAppManager):
            while True:
                await app_manager.run_agents_async()

                if run_once:
                    break
                log.info(
                    f"sleeping {polling_interval} seconds for app {app_manager.app.name}..."
                )
                await asyncio.sleep(polling_interval)

        for app_manager in self.managed_apps:
            await app_manager.start_polling_fluxnode_ips()
            tasks.append(asyncio.create_task(manage(app_manager)))

        await asyncio.gather(*tasks)


class FluxAppManager:
    def __init__(
        self,
        keeper: FluxKeeper,
        app: FluxApp,
    ):
        self.keeper = keeper
        self.app = app
        self.agents: list[RPCClient] = []
        self.network_state = {}
        # make network_state proper object and move last_row_build_somewhere there
        self.last_row_build: UnixTime = 0
        self.table_rows: list[str] = []
        # move to running tasks?
        self.fluxnode_sync_task: asyncio.Task | None = None
        self.running_tasks: dict[AgentId, list[asyncio.Task]] = {}
        self.sessions: dict[AgentId, list] = {}

        # This shouldn't be here, should be on the app
        if not self.app.signing_key and self.app.sign_connections:
            raise ValueError("Signing key must be provided if signing connections")

        self.register_extensions()

    def __iter__(self):
        yield from self.agents

    def __len__(self):
        return len(self.agents)

    @property
    def agent_ids(self) -> list:
        return [x.id for x in self.agents]

    @property
    def agent_ips(self) -> set:
        return set([x[1] for x in self.agent_ids])

    @property
    def primary_agents(self) -> filter:
        # return [x for x in self.agents if not x.is_proxied]
        return list(filter(lambda x: not x.is_proxied, self.agents))

    @staticmethod
    async def get_fluxnode_ips(app_name: str) -> list:
        url = f"https://api.runonflux.io/apps/location/{app_name}"
        timeout = aiohttp.ClientTimeout(connect=10)
        retries = 3

        data = {}

        # look at making session appwide
        async with aiohttp.ClientSession() as session:
            for n in range(retries):
                try:
                    async with session.get(url, timeout=timeout) as resp:
                        if resp.status in [429, 500, 502, 503, 504]:
                            log.error(f"bad response {resp.status}... retrying")
                            continue

                        data = await resp.json()
                        break

                except aiohttp.ClientConnectorError:
                    log.error(f"Unable to connect to {url}... retrying")
                    await asyncio.sleep(n)
                    continue

        node_ips = []
        if data.get("status", None) == "success":
            nodes = data.get("data")
            for node in nodes:
                ip = node["ip"].split(":")[0]
                node_ips.append(ip)
        else:
            log.error("Return status from Flux api was not successful for agent IPs")

        return node_ips

    @staticmethod
    def pad_to_width(target: list, width: int) -> Align:
        """Convience function to pre-pad a list of strings
        before passing to rich Align"""

        length = len(target)

        if length < width:
            target_filler = ["  "] * (width - length)
            target = [*target, *target_filler]

        return Align.center("".join(target))

    @staticmethod
    def add_symbol_to_list(
        subject: bool,
        target: list,
        insert_loc: int | None = None,
        colors: str = "red_green",
    ):
        """Append or insert different box colors based on truthiness"""
        green_box = "\U0001F7E9"  # True
        red_box = "\U0001F7E5"  # False
        yellow_box = "\U0001F7E8"  # True
        blue_box = "\U0001F7E6"  # False
        black_box = "\U00002B1B"  # None

        if colors == "red_green":
            output = green_box if subject else red_box

        elif colors == "blue_yellow":
            output = yellow_box if subject else blue_box
        else:
            output = "XX"

        # was getting false negatives with int
        if isinstance(insert_loc, int):
            target.insert(insert_loc, output)
        else:
            target.append(output)

    #   MODIFY APPEND METHOD FOR TRUE FALSE TO ADD SYMBOLS BINARY DEQUE (2 OPTIONS)

    def get_hit_counters(self) -> dict[str, HitCounter]:
        contact_states: dict[str, NodeContactState] = {
            k[1]: v.get("contact_state") for k, v in self.network_state.items()
        }

        return {k: v.hit_counter for k, v in contact_states.items() if v}

    def cache_or_build_table_rows(self, use_cache: bool) -> list:
        if not use_cache:
            return self.build_table_rows()
        else:
            hit_counters = self.get_hit_counters()

            latest = max(hit_counters.values(), key=lambda x: x.last_update, default=0)
            if latest:
                latest = latest.last_update

            if latest > self.last_row_build:
                self.last_row_build = latest
                return self.build_table_rows()
            else:
                return self.table_rows

    def build_table_rows(self) -> list:
        rows = []

        width = 15
        for target, hit_counter in self.get_hit_counters().items():
            one_sec_column = []
            one_min_column = []
            fifteen_min_column = []
            one_hour_column = []

            for ping in list(reversed(hit_counter.one_minute)):
                self.add_symbol_to_list(ping, one_min_column)

            for ping in list(reversed(hit_counter.fifteen_minute)):
                self.add_symbol_to_list(ping, fifteen_min_column)

            for ping in list(reversed(hit_counter.one_hour)):
                self.add_symbol_to_list(ping, one_hour_column)

            length = len(hit_counter.raw)
            match length // width:
                case x if x >= width:
                    four_sec_blocks = width
                    one_sec_blocks = 0
                case x if x < width:
                    if length <= width:
                        four_sec_blocks = 0
                        one_sec_blocks = width
                    else:
                        extra = length - width
                        # we have to also take into the account the realestate we loose
                        # from converting a one column to a 4 column. The net gain is 3
                        four_sec_blocks = extra // 3
                        if extra % 3 != 0:
                            four_sec_blocks += 1

                        one_sec_blocks = width - four_sec_blocks

            if not four_sec_blocks:
                for ping in reversed(hit_counter.raw):
                    self.add_symbol_to_list(ping, one_sec_column, colors="blue_yellow")
            else:
                fifo_pings = list(reversed(hit_counter.raw))
                # get the newest first
                for ping in fifo_pings[0:one_sec_blocks]:
                    self.add_symbol_to_list(ping, one_sec_column, colors="blue_yellow")

                # aggregate the one sec blocks to four sec blocks
                for i in range(four_sec_blocks):
                    # these are the oldest first
                    start = i * 4
                    blocks = list(itertools.islice(hit_counter.raw, start, start + 4))
                    last_4: bool = all(blocks)

                    self.add_symbol_to_list(last_4, one_sec_column, one_sec_blocks)

            rows.append(
                [
                    target,
                    self.pad_to_width(one_sec_column, width),
                    self.pad_to_width(one_min_column, width),
                    self.pad_to_width(fifteen_min_column, width),
                    self.pad_to_width(one_hour_column, width),
                ]
            )
        # cache
        self.table_rows = rows
        return rows

    def cancel_and_delete_tasks(self, agent_id: tuple):
        for task in self.running_tasks[agent_id]:
            task.cancel()

        del self.running_tasks[agent_id]

    def is_task_running_for_agent(self, agent_id: tuple, task_name: str) -> bool:
        found = False
        if agent_id in self.running_tasks:
            found = bool(
                next(
                    filter(
                        lambda x: x.get_name() == task_name,
                        self.running_tasks[agent_id],
                    ),
                    False,
                ),
            )
        return found

    async def start_polling_fluxnode_ips(self):
        """Idempotent polling of Fluxnodes"""
        if not self.fluxnode_sync_task:
            self.fluxnode_sync_task = asyncio.create_task(self.build_agents())

        while not self.agents:
            await asyncio.sleep(0.1)

    def add(self, agent: RPCClient):
        self.agents.append(agent)

    def remove(self, ip: str):
        remaining = []
        # there should only ever be one app running on a host
        for agent in self.agents:
            if agent.id[1] == ip:
                self.cancel_and_delete_tasks(agent.id)
            else:
                remaining.append(agent)

        self.agents = remaining

    async def disconnect_agent_by_id(self, id: AgentId):
        if agent := self.get_agent_by_id(id):
            await agent.transport.disconnect()

    def get_agent_by_id(self, id: AgentId):
        for agent in self.agents:
            if agent.id == id:
                return agent

    def get_by_socket(self, socket: tuple):
        for agent in self.agents:
            if not agent.connected:
                continue

            local = agent.transport.writer.get_extra_info("sockname")
            if local == socket:
                return agent

    def proxied_agents(self):
        # return filter(lambda x: x.is_proxied, self.agents)
        for agent in self.agents:
            if agent.is_proxied:
                yield agent

    def cleanup(self):
        self.fluxnode_sync_task.cancel()

    async def build_agents(self):
        while True:
            log.info("Fetching Fluxnode addresses...")
            fluxnode_ips = (
                self.app.fluxnode_ips
                if self.app.fluxnode_ips
                else await self.get_fluxnode_ips(self.app.name)
            )

            fluxnode_ips = set(fluxnode_ips)

            if not fluxnode_ips:  # error fetching ips from api
                log.error("Error fetching Fluxnode IPs... trying again in 30s")
                # try again soon
                await asyncio.sleep(30)
                continue

            missing = fluxnode_ips - self.agent_ips
            extra = self.agent_ips - fluxnode_ips

            for ip in extra:
                self.remove(ip)

            # this is stupid too. Theyre all getting the same auth provider anyway,
            # it only matters on the agent (they store stuff on the provider)
            auth_provider = None
            if self.app.sign_connections and self.app.signing_key:
                auth_provider = SignatureAuthProvider(key=self.app.signing_key)

            # this seems pretty broken
            if self.app.app_mode == AppMode.SINGLE_COMPONENT:
                component_name = self.app.get_component().name
            else:
                component_name = "fluxagent"

            try:
                for ip in missing:
                    agent_id = (self.app.name, ip, component_name)
                    self.network_state[agent_id] = {}

                    transport = EncryptedSocketClientTransport(
                        ip,
                        self.app.comms_port,
                        auth_provider=auth_provider,
                        proxy_target="",
                        on_pty_data_callback=self.keeper.gui.pty_output,
                        on_pty_closed_callback=self.keeper.gui.pty_closed,
                    )
                    flux_agent = RPCClient(JSONRPCProtocol(), transport, agent_id)
                    self.add(flux_agent)
                    log.info(f"Agent {flux_agent.id} added...")
            except Exception as e:
                print(repr(e))

            # if addresses were passed in, we don't need to loop
            if self.app.fluxnode_ips:
                break

            await asyncio.sleep(60)

    def register_extensions(self):
        self.app.extensions.add_method(self.get_all_agents_methods)
        self.app.extensions.add_method(self.poll_all_agents)

    def get_methods(self):
        """Returns methods available for the keeper to call"""
        return {k: v.__doc__ for k, v in self.app.extensions.method_map.items()}

    def get_all_agents_methods(self) -> dict:
        return self.keeper.loop.run_until_complete(self._get_agents_methods())

    @manage_transport
    async def get_agent_methods(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        methods = await agent_proxy.get_methods()

        return {agent.id: methods}

    async def _get_agents_methods(self) -> dict:
        """Queries every agent and returns a list describing what methods can be run on
        each agent"""
        tasks = []
        for agent in self.agents:
            task = asyncio.create_task(self.get_agent_methods(agent))
            tasks.append(task)

        results = await asyncio.gather(*tasks)
        return reduce(lambda a, b: {**a, **b}, results)

    @manage_transport
    async def get_state(self, agent: RPCClient):
        proxy = agent.get_proxy()
        self.network_state[agent.id].update(await proxy.get_state())

    @staticmethod
    def get_extra_objects(
        app_mode: AppMode,
        managed_object: FsEntryStateManager,
        local_hashes: dict[str, int],
        remote_hashes: dict[str, int],
    ) -> tuple[list[Path], int]:
        count = 0
        extras = []

        fake_root = managed_object.root()

        root_path = "/"
        for remote_name in remote_hashes:
            remote_name = Path(remote_name)

            target = str(fake_root / remote_name.relative_to(root_path))
            exists = local_hashes.get(target, None)

            if exists == None:
                count += 1
                if not extras:
                    extras.append(remote_name)

                extras = FsStateManager.filter_hierarchy(remote_name, extras)

        return (extras, count)

    @staticmethod
    def get_missing_or_modified_objects(
        app_mode: AppMode,
        managed_object: FsEntryStateManager,
        local_hashes: dict[str, int],
        remote_hashes: dict[str, int],
    ) -> tuple[list[Path], int, int]:
        # can't use zip here as we don't know remote lengths
        # set would work for filenames but not hashes
        # iterate hashes and find missing / modified objects

        missing = 0
        modified = 0
        candidates: list[Path] = []
        fake_root = managed_object.root()

        # if app_mode == AppMode.FILESERVER:
        #     remote_root = managed_object.remote_workdir
        # else:
        remote_root = "/"

        for local_path, local_crc in local_hashes.items():
            local_path = Path(local_path)

            remote_absolute = remote_root / local_path.relative_to(fake_root)

            # this should always be found, we asked for the hash.
            remote_crc = remote_hashes.get(str(remote_absolute), None)
            if remote_crc is None:  # 0 means empty file. Should just hash the filename
                missing += 1

            elif remote_crc != local_crc:
                modified += 1

            if remote_crc is None or remote_crc != local_crc:
                candidates.append(local_path)

        return (candidates, missing, modified)

    def resolve_object_deltas(
        self,
        managed_object: FsEntryStateManager,
        local_hashes: dict[str, int],
        remote_hashes: dict[str, int],
    ) -> tuple[list[Path], list[Path]]:
        candidates, missing, modified = self.get_missing_or_modified_objects(
            self.app.app_mode, managed_object, local_hashes, remote_hashes
        )

        extra_objects, unknown = self.get_extra_objects(
            self.app.app_mode, managed_object, local_hashes, remote_hashes
        )

        log.info(
            f"{missing} missing object(s), {modified} modified object(s) and {unknown} extra object(s)"
        )

        return (candidates, extra_objects)

    async def sync_remote_object(
        self,
        agent_proxy: RPCProxy,
        managed_object: FsEntryStateManager,
        object_fragments: list[Path] = [],
    ) -> dict:
        MAX_INBAND_FILESIZE = 1048576 * 50
        inband = False
        to_stream = []

        # this whole thing needs a refactor, gets called for both files and dirs
        # whole fragment thing is weird
        if object_fragments:
            remote_dir = managed_object.absolute_remote_path
            # think this is broken
            size = managed_object.concrete_fs.get_partial_size(object_fragments)
        else:
            remote_dir = managed_object.absolute_remote_dir
            object_fragments = [managed_object.concrete_fs.path]
            size = managed_object.concrete_fs.size

        # this logging is wrong. It implies that each object is the size which isn't correct
        # It's the aggregate size of all children if it's a dir and the size if it's a file
        #
        # log.info(
        #     f"Sending {bytes_to_human(size)} across {len(object_fragments)} object(s)"
        # )

        if size < MAX_INBAND_FILESIZE:
            inband = True

        for fs_entry in object_fragments:
            # this is dumb, but I'm tired. Ideally shouldn't be creating parent dirs
            # for files on remote end - they should get created explictily

            # this is true except for root dir, it's "fake root"
            # if syncing_root:
            #     # need to fake this
            #     abs_remote_path = str(remote_dir / managed_object.name)
            # else:
            #     print("RELATIVE", fs_entry, managed_object.local_path)
            if fs_entry != managed_object.local_parent / managed_object.name:
                relative = fs_entry.relative_to(
                    managed_object.local_parent / managed_object.name
                )
                abs_remote_path = str(remote_dir / relative)
            else:
                abs_remote_path = str(remote_dir / fs_entry.name)

            if fs_entry.is_dir():
                # Only need to do for empty dirs but currently doing on all dirs (wasteful as they will get created anyways)
                await agent_proxy.write_object(abs_remote_path, True, b"")
            elif fs_entry.is_file():
                # read whole file in one go as it's less than 50Mb
                if inband:
                    async with aiofiles.open(fs_entry, "rb") as f:
                        await agent_proxy.write_object(
                            abs_remote_path, False, await f.read()
                        )
                    continue
                else:
                    to_stream.append((fs_entry, abs_remote_path))
        if to_stream:
            transport = agent_proxy.get_transport()
            await transport.stream_files(to_stream)

    async def resolve_file_state(
        self, managed_object: FsEntryStateManager, agent_proxy: RPCProxy
    ):
        log.info(
            f"File {managed_object.name} with size {bytes_to_human(managed_object.concrete_fs.size)} is about to be transferred"
        )
        # this seems a bit strange but writing directory uses the same interface
        # and they don't know who the file names are, the just have the associated
        # managed_object

        # what are we passing in here?
        await self.sync_remote_object(agent_proxy, managed_object)

        managed_object.in_sync = True
        managed_object.remote_object_exists = True

    async def resolve_directory_state(
        self,
        component: FluxComponent,
        managed_object: FsEntryStateManager,
        agent_proxy: RPCProxy,
    ) -> list:
        remote_path = str(managed_object.absolute_remote_path)

        # if it doesn't exist - no point getting child hashes
        if managed_object.remote_crc == 0:
            await self.sync_remote_object(agent_proxy, managed_object)
            return []

        # this is different from the global get_all_object_hashes - this adds
        # all the hashes together, get_directory_hashes keeps them seperate
        remote_hashes = await agent_proxy.get_directory_hashes(remote_path)
        # these are absolute
        local_hashes = managed_object.concrete_fs.get_directory_hashes(
            name=managed_object.name
        )

        # these are in remote absolute form
        object_fragments, objects_to_remove = self.resolve_object_deltas(
            managed_object, local_hashes, remote_hashes
        )

        if (
            managed_object.remit.sync_strategy == SyncStrategy.STRICT
            and objects_to_remove
        ):
            # we need to remove extra objects
            # ToDo: sort serialization so you can pass in paths etc
            to_delete = [str(x) for x in objects_to_remove]
            await agent_proxy.remove_objects(to_delete)
            log.info(
                f"Sync strategy set to {SyncStrategy.STRICT.name} for {managed_object.name}, deleting extra objects: {to_delete}"
            )
        elif SyncStrategy.ALLOW_ADDS:
            managed_object.validated_remote_crc = managed_object.remote_crc

        log.info(
            f"Deltas resolved... {len(object_fragments)} object(s) need to be resynced"
        )

        if object_fragments:
            await self.sync_remote_object(agent_proxy, managed_object, object_fragments)
            component.state_manager.set_syncronized(object_fragments)

        managed_object.in_sync = True
        managed_object.remote_object_exists = True

        return object_fragments

    @manage_transport
    async def load_manifest(self, agent: RPCClient):
        """This is solely for the fileserver"""

        if not agent.transport.auth_provider:
            log.warn("Agent not using auth, unable to sign manifest... skipping")
            return

        component = self.app.get_component(agent.id[2])
        managed_object = component.state_manager.get_object_by_remote_path(WWW_ROOT)
        fileserver_hash = managed_object.local_crc
        # this only works if we're signing messages
        sig = agent.transport.auth_provider.sign_message(str(fileserver_hash))
        # manifest = managed_object.concrete_fs.decendants()
        agent_proxy = agent.get_proxy()
        await agent_proxy.load_manifest(fileserver_hash, sig)

    @manage_transport
    async def sync_objects(self, agent: RPCClient):
        log.debug(f"Contacting Agent {agent.id} to check if files required")
        # ToDo: fix formatting nightmare between local / common / remote
        component = self.app.get_component(agent.id[2])

        if not component:
            # each component must be specified
            log.warn(
                f"No config found for component {agent.id[2]}, this component will only get globally specified files"
            )
            return

        remote_paths = component.state_manager.absolute_remote_paths()

        agent_proxy = agent.get_proxy()

        # start = time.monotonic()
        remote_fs_objects = await agent_proxy.get_all_object_hashes(remote_paths)
        # elapsed = time.monotonic() - start
        # log.info(f"Time to get hashes: {round(elapsed, 4)}")

        log.debug(f"Agent {agent.id} remote file CRCs: {remote_fs_objects}")

        if not remote_fs_objects:
            log.warn(f"No objects to sync for {agent.id} specified... skipping!")
            return

        fixed_objects = []

        for remote_fs_object in remote_fs_objects:
            # this is broken. If we're a dir, any children have already been fixed. Don't
            # need to resolve them too. So need to keep track of the parent.
            remote_path = Path(remote_fs_object["name"])
            managed_object = component.state_manager.get_object_by_remote_path(
                remote_path
            )

            if managed_object.local_path in fixed_objects:
                managed_object.remote_crc = managed_object.local_crc
                return

            if not managed_object:
                log.warn(f"managed object: {remote_path} not found in component config")
                return

            managed_object.remote_crc = remote_fs_object["crc32"]
            # this just crc's the local object, I had this disabled so only validate files
            # on boot but meh - disabled again, it's fucking slow and blocking my shit, tried in a thread
            # but the expensive of shunting to a thread is probably too much. I should probably split this whole
            # function from sync / async and run in thread
            # await asyncio.to_thread(managed_object.validate_local_object)
            managed_object.compare_objects()

            if not managed_object.local_object_exists:
                log.warn(
                    f"Remote object exists but local doesn't: {managed_object.local_path}. Local workdir: {managed_object.local_workdir}"
                )
                continue

            if (
                managed_object.remit.sync_strategy == SyncStrategy.ALLOW_ADDS
                and managed_object.concrete_fs.storable
                and not managed_object.in_sync
                and managed_object.validated_remote_crc == managed_object.remote_crc
            ):
                continue

            if (
                managed_object.remit.sync_strategy == SyncStrategy.ENSURE_CREATED
                # and managed_object.concrete_fs.storable
                and managed_object.remote_object_exists
            ):
                continue

            if (
                managed_object.concrete_fs.storable
                and managed_object.concrete_fs.empty
                and managed_object.local_crc == managed_object.remote_crc
            ):
                continue

            if not managed_object.in_sync and managed_object.concrete_fs.storable:
                # so here we need to figure any children dirs and set them manually to in_sync
                fixed = await self.resolve_directory_state(
                    component, managed_object, agent_proxy
                )
                fixed_objects.extend(fixed)

            if not managed_object.in_sync and not managed_object.concrete_fs.storable:
                await self.resolve_file_state(managed_object, agent_proxy)

        # log.info(f"Time to sync objects: {time.monotonic() - start}")

    def poll_all_agents(self):
        self.keeper.loop.run_until_complete(self.run_agent_tasks())

    @manage_transport
    async def load_agent_plugins(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        await agent_proxy.load_plugins()

    @manage_transport
    async def enroll_agent(self, agent: RPCClient):
        log.info(f"Enrolling agent {agent.id}")
        proxy = agent.get_proxy()
        res = await proxy.generate_csr()
        csr_bytes = res.get("csr")

        csr = cryptography.x509.load_pem_x509_csr(csr_bytes)
        hostname = csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value

        try:
            cert = self.keeper.ca.load_certificate(hostname)
            self.keeper.ca.revoke_certificate(hostname)
        except OwnCAInvalidCertificate:
            pass
        finally:
            # ToDo: there has to be a better way (don't delete cert)
            # start using CRL? Do all nodes need CRL - probably
            shutil.rmtree(f"ca/certs/{hostname}", ignore_errors=True)
            cert = self.keeper.ca.sign_csr(csr, csr.public_key())

        # This triggers agent to update registrar, it should probably
        # be it's own action
        await proxy.install_cert(cert.cert_bytes)
        await proxy.install_ca_cert(self.keeper.ca.cert_bytes)

    def get_task_state(self) -> dict:
        all_agents = {k.id: [] for k in list(self.agents)}
        all_tasks = {
            k: [x.get_name() for x in v] for k, v in self.running_tasks.items()
        }
        all_agents.update(all_tasks)
        # build data model for ping stuff
        # this function should just add to network_state

    @staticmethod
    async def wait_for_events(events: list[asyncio.Event]):
        """Await each event, this ensures all events have completed"""
        for event in events:
            await event.wait()

    async def ping_all_nodes(
        self, down_callback: Callable, up_callback: Callable, interval: int = 1
    ):
        sync_events: list[asyncio.Event] = []

        for agent in self.agents:
            sync_event = asyncio.Event()
            sync_events.append(sync_event)
            agent_tasks = self.running_tasks.get(agent.id, [])

            ping_task = next(
                filter(lambda x: x.get_name() == "ping", agent_tasks), None
            )
            if not ping_task:
                t = asyncio.create_task(
                    self.ping_pong(
                        agent,
                        interval=interval,
                        down_callback=down_callback,
                        up_callback=up_callback,
                        sync_event=sync_event,
                    ),
                    name="ping",
                )
                if agent.id not in self.running_tasks:
                    self.running_tasks[agent.id] = [t]
                else:
                    self.running_tasks[agent.id].append(t)

        await self.wait_for_events(sync_events)

    @manage_transport(exclusive=True)
    async def ping_pong(
        self,
        agent: RPCClient,
        interval: int,
        down_callback: Callable,
        up_callback: Callable,
        sync_event: asyncio.Event,
    ):
        """Ping node once every interval seconds. This interval includes the RTT of the ping,
        if the RTT exceeds the interval, it is pinged again immediately. If the nodes
        missed 3 pings (intervals) or doesn't respond with the correct PONG - the user provided
        callback is called and we stop pinging"""

        print("PING PONG BABY")
        # we are saying we have connected (decorator has run)
        sync_event.set()

        state = NodeContactState()
        self.network_state[agent.id].update({"contact_state": state})
        agent_proxy = agent.get_proxy(exclusive=True)

        async def run_callback(callback) -> bool:
            # this_task = next(
            #     filter(
            #         lambda x: x.get_name() == "ping",
            #         self.running_tasks[agent.id],
            #     )
            # )

            # self.running_tasks[agent.id].remove(this_task)

            if asyncio.iscoroutinefunction(callback):
                await callback(agent.id)
            else:
                callback(agent.id)

            # this_task.cancel()

        async def ping_forever():
            # logic
            DOWN_NODE_INTERVAL = interval * 10
            QUARANTINE_TIMER = 300
            # if 3 missed in a row... you're out of here
            # if 6 missed in a one minute period... you're out of here
            consecutive_misses = 0
            while True:
                # switch back to monotonic for those gainz
                start = time.perf_counter()
                try:
                    # we need shield here, if we don't and we timeout, the task gets cancelled,
                    # which stops the transport pulling the message off the queue; depending on where we
                    # # hit the timeout. So we could have an extra
                    # message in there, which gets returned immediately, making it look like the
                    # elapsed time is very, very, short
                    # However, what happens if we get 3 timeouts, do we need to cancel
                    # the tasks
                    agent_proxy.set_timeout(interval)
                    # resp = await asyncio.wait_for(
                    #     asyncio.shield(agent_proxy.ping()), timeout=interval
                    # )
                    resp = await agent_proxy.ping()
                except asyncio.TimeoutError as e:
                    state.update_hit_counter(False)
                    consecutive_misses += 1
                    log.error(
                        f"Error pinging remote: {agent.id}. Consecutive_misses: {consecutive_misses}"
                    )
                    if consecutive_misses >= 10:
                        consecutive_misses = 0
                        await agent.transport.disconnect()
                        try:
                            await agent.transport.ensure_connected()
                        except Exception as e:
                            print("orangutan")
                            print(repr(e))
                        continue

                    elif consecutive_misses >= 4 or state.one_minute_miss_count >= 8:
                        if state.active:
                            state.transitions.append(StateTransition(False))
                            await run_callback(down_callback)

                    else:
                        state.increment_counters()

                except (ConnectionResetError, BrokenPipeError) as e:
                    log.error(f"Agent: {agent.id} disconnected with E: {repr(e)}")
                    state.increment_counters()
                    consecutive_misses = 0
                    await agent.transport.disconnect()
                    if state.active:
                        state.transitions.append(StateTransition(False))
                        await run_callback(down_callback)
                    try:
                        await agent.transport.ensure_connected()
                    except Exception as e:
                        print("giraffe")
                        print(repr(e))

                except Exception as e:
                    print(f"Exiting ping forever: {repr(e)}")
                    exit(1)
                else:
                    rtt = time.monotonic() - start
                    try:
                        state.update_rtt(rtt)
                        state.update_hit_counter(True)
                    except Exception as e:
                        print(repr(e))

                    consecutive_misses = 0
                    if resp != "PONG":
                        log.warning(f"Received incorrect ping response: {resp}")

                    if not state.active and not state.in_quarantine:
                        log.info(
                            f"Node resumed... putting into quarantine for 300 seconds"
                        )
                        state.transitions.append(StateTransition(True))
                        state.quarantine()

                    elif state.in_quarantine:
                        if (
                            start - state.quarantine_timer > QUARANTINE_TIMER
                            and state.one_minute_miss_count < 6
                        ):
                            log.info(f"Node has passed quarantine tests... reinstating")
                            state.dequarantine()
                            await run_callback(up_callback)
                finally:
                    state.total_count += 1
                    elapsed = time.perf_counter() - start
                    # if state.total_count % 10 == 0:
                    #     print(
                    #         f"{agent.id} count: {state.total_count} elapsed: {elapsed}"
                    #     )

                    if not state.active and not state.in_quarantine:
                        timer = DOWN_NODE_INTERVAL
                    else:
                        timer = interval
                    to_sleep = max(0, timer - elapsed)
                    # print(f"To sleep: {to_sleep}")
                    await asyncio.sleep(to_sleep)

        print("about to ping forever")
        await ping_forever()
        # pprint(asyncio.all_tasks())

    @manage_transport
    async def set_mode(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        resp = await agent_proxy.set_mode(self.app.app_mode.value)

    # @manage_transport
    # async def enable_fileserver(self, agent: RPCClient):
    #     agent_proxy = agent.get_proxy()
    #     resp = await agent_proxy.enable_registrar_fileserver()

    @manage_transport
    async def get_agents_state(self, agent: RPCClient) -> str:
        agent_proxy = agent.get_proxy()
        resp = await agent_proxy.get_container_state()

        return resp

    @manage_transport
    async def enroll_subordinates(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        resp = await agent_proxy.get_subagents()

        sub_names = [k for k in resp["sub_agents"]]
        log.info(f"Agent {agent.id} has the following subordinates: {sub_names}")
        address = agent.transport.address

        for target, payload in resp.get("sub_agents").items():
            role = payload.get("role")  # not implemented yet
            enrolled = payload.get("enrolled")

            # ToDo: check if already enrolled, may have rebooted

            if not enrolled:
                flux_agent = self.create_agent(address, target)
                await self.enroll_agent(flux_agent, True, True)
                self.add(flux_agent)

    def create_agent(
        self,
        address: str,
        proxy_target: str | None = None,
        auth_provider: SignatureAuthProvider | None = None,
    ) -> RPCClient:
        transport = EncryptedSocketClientTransport(
            address,
            self.app.comms_port,
            auth_provider=auth_provider,
            proxy_target=proxy_target,
            proxy_port=self.app.comms_port,
            proxy_ssl=False,
            cert=self.keeper.cert,
            key=self.keeper.key,
            ca=self.keeper.ca_cert,
            on_pty_data_callback=self.keeper.gui.pty_output,
            on_pty_closed_callback=self.keeper.gui.pty_closed,
        )
        flux_agent = RPCClient(
            JSONRPCProtocol(), transport, (self.app.name, address, proxy_target)
        )
        flux_agent.transport: EncryptedSocketClientTransport

        return flux_agent

    def set_default_tasks(self) -> list[Callable]:
        tasks = [
            self.build_task("sync_objects"),
            self.build_task("set_mode"),
            self.build_task("get_state"),
        ]

        if self.app.app_mode != AppMode.FILESERVER:
            tasks.insert(0, FluxTask("enroll_subordinates"))

        if self.app.app_mode == AppMode.FILESERVER:
            tasks.insert(2, FluxTask("load_manifest"))

        return tasks

    def build_task(
        self, name: str, args: list = [], kwargs: dict = {}
    ) -> FluxTask | None:
        try:
            func = getattr(self, name)
        except AttributeError:
            log.warn(f"Task {name} not found, skipping")
            return

        return FluxTask(name=name, args=args, kwargs=kwargs, func=func)

    async def reachable_via_proxy(self, proxy_ip: str, target_ip: str) -> bool:
        agent = self.create_agent(proxy_ip, target_ip)

        await agent.transport.connect()

        if not agent.transport.connected:
            return False

        await agent.transport.disconnect()

        return True

    @asynccontextmanager
    async def agent_sessions(self, agents: list[tuple] = []):
        if agents:
            agents = [self.get_agent_by_id(x) for x in agents]
            agents = list(filter(None, agents))
        else:
            agents = self.agents

        for agent in agents:
            await agent.transport.session.start()

        yield

        for agent in agents:
            await agent.transport.session.end()

    async def task_runner(self, agent: RPCClient, tasks: list[FluxTask]) -> dict:
        # con_kwargs = {"in_session": True}
        results = {}

        for task in tasks:
            if not task.func:
                continue

            res = await task.func(agent, *task.args, **task.kwargs, in_session=True)
            results[task.func.__name__] = res

        return {agent.id: results}

    async def run_session_tasks(
        self,
        tasks: list[FluxTask] = [],
        targets: dict[AgentId, list[FluxTask]] = {},
    ) -> dict[tuple, dict]:
        if not self.agents:
            log.info("No agents found... nothing to do")
            return

        if not tasks and not targets:
            tasks = self.set_default_tasks()

        agents = []
        for target, agent_tasks in targets.items():
            if agent := self.get_agent_by_id(target):
                agents.append((agent, agent_tasks))
            else:
                log.warn(f"Target {target} not found... nothing to do")

        if not agents:
            agents = [(x, tasks) for x in self.agents]

        # log.info(f"Agent session tasks: {pretty_repr(agents)}")

        coroutines = []
        for agent, tasks in agents:
            t = asyncio.create_task(self.task_runner(agent, tasks))
            coroutines.append(t)

        try:
            results = await asyncio.gather(*coroutines)
        except FluxVaultKeyError as e:
            log.error(f"Exception from gather tasks: {repr(e)}")
            if self.keeper.gui:
                await self.keeper.gui.set_toast(repr(e))

        # Going for speed gains... this takes hardly any time but trying for improvements
        # if self.keeper.gui:
        #     await self.keeper.gui.app_state_update(self.app.name, self.network_state)

        results = list(filter(None, results))
        results = reduce(lambda a, b: {**a, **b}, results)

        return results

    async def run_agents_async(
        self,
        tasks: list[FluxTask] = [],
        stay_connected: bool = False,
        already_connected: bool = False,
        targets: dict[AgentId, list[FluxTask]] = {},
        async_tasks: bool = False,  # NotImplemented
    ) -> dict[tuple, dict]:
        if not self.agents:
            log.info("No agents found... nothing to do")
            return

        if not tasks and not targets:
            tasks = self.set_default_tasks()

        coroutines = []
        print("stay connected", stay_connected)
        print("already connected", already_connected)

        async def task_runner(agent: RPCClient, tasks: list[FluxTask]) -> dict:
            length = len(tasks)
            results = {}

            for index, task in enumerate(tasks):
                # first task connects, last task disconnects

                if agent.transport.failed_on == NO_SOCKET:
                    # this is super dirty
                    agent.transport.failed_on = ""
                    # means we skip all tasks in this run but can then run
                    # again next time
                    break

                if not task.func:
                    continue

                connect = False
                disconnect = False
                if index == 0 and not already_connected:
                    connect = True
                if index + 1 == length and not stay_connected:
                    disconnect = True

                con_kwargs = {
                    "connect": connect,
                    "disconnect": disconnect,
                    "in_session": False,
                }

                # doesn't this mean you have to use @manage_transport???
                res = await task.func(agent, *task.args, **task.kwargs, **con_kwargs)
                results[task.func.__name__] = res

            # agent.id is a tuple of app.name, ip, component_name
            return {agent.id: results}

        log.info(f"Agent session tasks: {pretty_repr(targets)}")

        agents = []
        for target, agent_tasks in targets.items():
            # if not seesion.started ??? connect? Should a session encapsulate
            # an agent?
            if agent := self.get_agent_by_id(target):
                agents.append((agent, agent_tasks))
            else:
                log.warn(f"Target {target} not found... nothing to do")

        if not agents:
            agents = [(x, tasks) for x in self.agents]

        for agent, tasks in agents:
            t = asyncio.create_task(task_runner(agent, tasks))
            coroutines.append(t)

        try:
            results = await asyncio.gather(*coroutines)
        except FluxVaultKeyError as e:
            log.error(f"Exception from gather tasks: {repr(e)}")
            if self.keeper.gui:
                await self.keeper.gui.set_toast(repr(e))

        if self.keeper.gui:
            await self.keeper.gui.app_state_update(self.app.name, self.network_state)

        results = list(filter(None, results))
        results = reduce(lambda a, b: {**a, **b}, results)

        return results

    def __getattr__(self, name: str) -> Callable:
        try:
            func = self.app.extensions.get_method(name)
        except MethodNotFoundError as e:
            raise AttributeError(f"Method does not exist: {e}")

        if func.pass_context:
            context = FluxVaultContext(self.agents, self.keeper.ca)
            name = func.__name__
            func = functools.partial(func, context)
            func.__name__ = name

        return func



# so you don't need to quote foward lookahead typing
from __future__ import annotations

import binascii
from collections.abc import Callable

# from typing import TYPE_CHECKING
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import BinaryIO, Generator
import time
import asyncio

import aiofiles

from fluxvault.helpers import RemoteStateDirective
from fluxvault.log import log

# if TYPE_CHECKING:
#     from fluxvault.fluxapp import RemoteStateDirective

BUFFERSIZE = 1048576 * 50

### BUILD NOTES ###

# File or dir
# └── = last file no more dirs
# ├── = last dir no files

# Depth
# "│   "
# "│   │   "
# "│   │   "
# "│   │   │   "

# if last dir at depth:
# └── = dir
# direct children files have "    " instead of "|   " for parent

# any child depths (greater than current depth) have
# "    " instead of "│   " at left

# "    " = spacer
# "|   " = ancestor line

# is parent last sibling? Yes Then spacer at parent depth -1

# slots. Each height has depth -1 slots

# slots are either spacers or ancestors

# eg

# spacer ancestor
# spacer ancestor ancestor spacer

### /BUILD NOTES ###


class FileTooLargeError(Exception):
    """"""

    ...


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def bytes_to_human(num, suffix="B"):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class FsType(Enum):
    DIRECTORY = 1
    FILE = 2
    UNKNOWN = 3


FORMATS: dict[str, Callable] = {
    "root_path": lambda size, path: f"{bytes_to_human(size)} {bcolors.OKBLUE}{path}{bcolors.ENDC}",
    "root_path_last_file": lambda prefix, size, path: f"{prefix}└── {bytes_to_human(size)} {path.name}",
    "root_path_not_last_file": lambda prefix, size, path: f"{prefix}├── {bytes_to_human(size)} {path.name}",
    "dir_terminal": lambda prefix, size, path: f"{prefix}└── {bytes_to_human(size)} {bcolors.OKBLUE}{path.name}{bcolors.ENDC}",
    "dir_not_terminal": lambda prefix, size, path: f"{prefix}├── {bytes_to_human(size)} {bcolors.OKBLUE}{path.name}{bcolors.ENDC}",
    "file_terminal_dir_last_file": lambda prefix, size, path: f"{prefix}{'    '}└── {bytes_to_human(size)} {path.name}",
    "file_terminal_dir_not_last_file": lambda prefix, size, path: f"{prefix}{'    '}├── {bytes_to_human(size)} {path.name}",
    "file_not_terminal_dir_last_file": lambda prefix, size, path: f"{prefix}{'│   '}└── {bytes_to_human(size)} {path.name}",
    "file_not_terminal_dir_not_last_file": lambda prefix, size, path: f"{prefix}{'│   '}├── {bytes_to_human(size)} {path.name}",
}


@dataclass
class CrcCacheItem:
    crc: int
    last_update: float = field(default_factory=time.monotonic)


@dataclass
class CrcCache:
    cache_items: dict[str, CrcCacheItem] = field(default_factory=dict)

    def get(self, name) -> int | None:
        if cache_item := self.cache_items.get(name):
            return cache_item.crc

    async def add(self, name, crc):
        self.cache_items[name] = CrcCacheItem(crc)
        # should be storing this somewhere
        asyncio.create_task(self.wait_and_remove(40, name))

    async def wait_and_remove(self, timeout, name):
        await asyncio.sleep(timeout)
        self.remove(name)

    def remove(self, name):
        try:
            del self.cache_items[name]
        except ValueError:
            pass


@dataclass
class ConcreteFsEntry:
    path: Path
    parent: ConcreteFsEntry | None = None
    children: list[ConcreteFsEntry] = field(default_factory=list)
    children_index: dict[str, int] = field(default_factory=dict)
    depth: int = 0
    last_modified: int = 0
    fs_type: FsType = FsType.UNKNOWN
    size: int = 0
    fh: BinaryIO | None = None
    crc_cache: CrcCache = field(default_factory=CrcCache)

    def __str__(self):
        """"""
        ancestor_symbols = {
            True: "    ",
            False: "│   ",
        }

        # object_symbols = {"not_last": "├── ", "last": "└── "}
        # file_symbols = {"not_last_dir": "│   ", "last_dir": "    "}

        files = [x for x in self.children if x.fs_type == FsType.FILE]
        dirs = [x for x in self.children if x.fs_type == FsType.DIRECTORY]
        is_last_sibling_dir = False

        ancestors = self.parents()

        slots = []
        for index, ancestor in enumerate(ancestors):
            if index == len(ancestors) - 1:
                target = self
            else:
                target = ancestors[index + 1]
            is_last_sibling_dir = ancestor.last_sibling_dir(target)
            slots.append(is_last_sibling_dir)

        terminal_dir = slots.pop() if slots else False
        prefix = "".join([ancestor_symbols[x] for x in slots])

        if self.depth == 0:
            dir_type = "root_path"
            params = [self.size, self.path]
        elif terminal_dir:
            dir_type = "dir_terminal"
            params = [prefix, self.size, self.path]
        else:
            dir_type = "dir_not_terminal"
            params = [prefix, self.size, self.path]

        directory = self.formatter(dir_type)(*params)

        formatted_files = []
        for index, file in enumerate(files):
            params = [prefix, file.size, file.path]
            if self.depth == 0:
                file_type = "root_path_not_last_file"
                if index + 1 == len(files) and not dirs:
                    file_type = "root_path_last_file"
            elif terminal_dir:
                file_type = "file_terminal_dir_not_last_file"
                if index + 1 == len(files) and not dirs:
                    file_type = "file_terminal_dir_last_file"
            else:
                file_type = "file_not_terminal_dir_not_last_file"
                if index + 1 == len(files) and not dirs:
                    file_type = "file_not_terminal_dir_last_file"
            formatted_files.append(self.formatter(file_type)(*params))

        # the str(d) list comp is recursive. I.e. - it calls this function
        return "\n".join([directory, *formatted_files, *[str(d) for d in dirs]])

    @classmethod
    def log_objects_in_dir(cls, dir: Path):
        if not dir.is_dir():
            return

        for child in dir.iterdir():
            log.info(f"Existing fake root object: {child} (before linking)")

    @classmethod
    def contains(cls, name: Path, dir: Path) -> bool:
        """If the directory contains the file name"""
        return (dir / name).exists()

    @classmethod
    def are_files_in_dir(cls, dir: Path) -> bool:
        if not dir.is_dir():
            return False

        return any(x.is_file() for x in dir.iterdir())

    @classmethod
    def entries_in_dir(cls, dir: Path) -> Generator[Path, None, None]:
        if not dir.is_dir:
            return []

        yield from dir.iterdir()

    @classmethod
    def find_first_path(cls, target: Path, locations: list[Path]) -> Path | None:
        """Pass in a path, it will return it if it exists, and if not,
        it will search in in paths (dirs) to find a name match. It found
        returns the full path"""

        first_match_dir = next(
            filter(lambda x: ConcreteFsEntry.contains(target, x), locations)
        )

        return first_match_dir

    @classmethod
    def build_tree(cls, base_path: Path, depth: int = 0) -> ConcreteFsEntry:
        children: list[ConcreteFsEntry] = []
        children_index = {}

        if not base_path.exists():
            return ConcreteFsEntry(base_path)

        if base_path.is_dir():
            for child in sorted(base_path.iterdir()):
                if child.is_dir():
                    fs_entry = ConcreteFsEntry.build_tree(child, depth + 1)
                elif child.is_file():
                    fs_entry = ConcreteFsEntry(
                        child, None, [], depth + 1, fs_type=FsType.FILE
                    )
                else:
                    raise Exception("FUCKED")

                children.append(fs_entry)

        parent = ConcreteFsEntry(
            base_path,
            parent=None,
            children=children,
            depth=depth,
            fs_type=FsType.DIRECTORY,
        )

        for index, child in enumerate(children):
            key = child.path.relative_to(parent.path)
            children_index[key] = index
            child.parent = parent

        parent.children_index.update(children_index)

        return parent

    @property
    def exists(self) -> bool:
        return self.path.exists()

    @property
    def empty(self) -> bool:
        """Are we empty"""
        return self.get_size == 0

    @property
    def child_dirs(self) -> bool:
        """Lets the caller know if this object has any children dirs"""
        return bool(len([x for x in self.children if x.path.is_dir()]))

    @property
    def readable(self) -> bool:
        """If this object can be called by read()"""
        return self.path.is_file()

    @property
    def storable(self) -> bool:
        """If this object can be used to store files"""
        return self.path.is_dir()

    @property
    def sibling_dirs(self) -> list:
        # this is wrong. use parent (it works but we should use our own interface)
        return self.path.is_dir() and any(
            [x for x in self.path.parent.iterdir() if x.is_dir() and x != self.path]
        )

    @staticmethod
    def formatter(section) -> Callable:
        return FORMATS[section]

    def get_child(self, target: Path) -> ConcreteFsEntry | None:
        relative_path = target.relative_to(self.path)

        if len(relative_path.parts) > 1:
            child_index = self.children_index.get(Path(relative_path.parts[0]))
            return self.children[child_index].get_child(target)

        if child_index := self.children_index.get(relative_path, None):
            return self.children[child_index]

    def get_partial_size(self, fragments: list[Path]) -> int:
        """Only makes sense to be called from a directory. However gets called
        recursively so will be for a file too"""

        sizes = []
        for target_path in fragments:
            if target_path == self.path:
                continue

            if target := self.get_child(target_path):
                sizes.append(target.size)
                continue

        return sum(sizes)

    def get_size(self) -> int:
        match self.fs_type:
            case FsType.FILE:
                return self.path.stat().st_size
            case FsType.DIRECTORY:
                return sum(
                    f.stat().st_size for f in self.path.glob("**/*") if f.is_file()
                )

    def root(self) -> ConcreteFsEntry:
        if self.depth == 0:
            return self

        ancestor_count = self.depth
        parent = None

        while ancestor_count > 0:
            ancestor_count -= 1
            if not parent:
                parent = self.parent
            else:
                parent = parent.parent

        return parent

    def parents(self) -> list[ConcreteFsEntry]:
        """Get all ancestors up the file tree, finishing at root"""
        parents: list[ConcreteFsEntry] = []
        ancestor_count = self.depth

        while ancestor_count > 0:
            ancestor_count -= 1
            if not parents:
                parents.append(self.parent)
            else:
                parents.insert(0, parents[0].parent)
        return parents

    def decendants(self) -> list[Path]:
        """Returns all child paths down the tree"""
        if self.path.is_dir():
            return [f for f in self.path.glob("**/*")]
        else:
            return []

    def last_sibling_dir(self, child: ConcreteFsEntry) -> bool:
        """Called from a child to a parent; Finds out if child is the last
        directory in parent's list of children"""
        if len(self.children):
            last = [x for x in self.children if x.fs_type == FsType.DIRECTORY][-1]
            return last.path == child.path
        return False

    def last_sibling(self, child: ConcreteFsEntry) -> bool:
        if len(self.children):
            last = self.children[-1]
            return last.path == child.path
        return False

    def recurse(self) -> Generator[ConcreteFsEntry, None, None]:
        yield self

        for child in self.children:
            yield from child.recurse()

    def realize(self):
        """Will populate FsEntry (and children) with live file details"""
        for child in self.children:
            child.realize()

        if self.readable:
            self.fs_type = FsType.FILE
            stat = self.path.stat()
            self.size = stat.st_size
            self.last_modified = stat.st_mtime

        elif self.storable:
            self.fs_type = FsType.DIRECTORY

            files_size = sum(f.size for f in self.children if f.readable)
            dirs_size = sum(d.size for d in self.children if d.storable)

            self.size = files_size + dirs_size
            self.last_modified = self.path.stat().st_mtime

    async def _reader(self, chunk_size: int) -> Generator[bytes, None, None]:
        if not self.fh:
            self.fh = await aiofiles.open(self.path, "rb").__aenter__()
        yield await self.fh.read(chunk_size)

    async def read(self, chunk_size: int | None = None) -> bytes:
        f"""Reads underlying file if entry is under {bytes_to_human(BUFFERSIZE)}, or
        reads up to chunk_size bytes. Uses a generator so file bytes aren't stored in buffer.
        """
        if chunk_size == None:  # reading until eof
            if self.size > BUFFERSIZE:
                raise FileTooLargeError(str(self.path))
        if not self.readable:
            raise FileNotFoundError(str(self.path))

        return await anext(self._reader(chunk_size))

    async def close(self):
        await self.fh.close()
        self.fh = None

    ### CRC OPERATIONS

    def crc_file(self, filename: Path, crc: int) -> int:
        # if cache_crc := self.crc_cache.get(str(filename)):
        #     return cache_crc

        crc = binascii.crc32(filename.name.encode(), crc)

        with open(filename, "rb") as f:
            for chunk in iter(lambda: f.read(1024 * 128), b""):
                crc = binascii.crc32(chunk, crc)

        # loop = asyncio.get_event_loop()
        # loop.run_until_complete(self.crc_cache.add(str(filename), crc))

        return crc

    def crc_directory(self, directory: Path, crc: int, name: str = "") -> int:
        # if cache_crc := self.crc_cache.get(str(directory)):
        #     return cache_crc

        dir_name = name if name else directory.name
        crc = binascii.crc32(dir_name.encode(), crc)

        for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
            crc = binascii.crc32(path.name.encode(), crc)

            if path.is_file():
                crc = self.crc_file(path, crc)
            elif path.is_dir():
                crc = self.crc_directory(path, crc)

        # loop = asyncio.get_event_loop()
        # loop.run_until_complete(self.crc_cache.add(str(directory), crc))

        return crc

    def crc32(self, name: str = "") -> int:
        if self.fs_type == FsType.DIRECTORY:
            return self.crc_directory(self.path, 0, name=name)
        elif self.fs_type == FsType.FILE:
            return self.crc_file(self.path, 0)

    def get_file_hash(self, file: Path):
        crc = self.crc_file(file, 0)
        return {str(file): crc}

    def get_directory_hashes(
        self, dir: Path | None = None, name: str | None = None
    ) -> dict[str, int]:
        """Returns a dictionary of every fs object and it's corresponding crc32"""
        hashes = {}

        # fix all this... needs to hash the filename too, otherwise
        # we get the wrong hash if new empty file added
        if not dir and not self.storable:
            return hashes

        p = dir if dir else self.path

        dir_name = name if name else p.name

        crc = binascii.crc32(dir_name.encode())
        hashes.update({str(p): crc})

        for path in sorted(p.iterdir(), key=lambda p: str(p).lower()):
            if path.is_dir():
                hashes.update(self.get_directory_hashes(path))

            elif path.is_file():
                hashes.update(self.get_file_hash(path))

        return hashes


@dataclass
class FsEntryStateManager:
    """Wrapper around ConcreteFsEntry. An abstract Filesystem. Syncs state from local to remote
    objects. From Keeper point of view, the StreamWriter for the remote system
    is the filesystem"""

    name: Path  # dir or file
    remit: RemoteStateDirective
    local_parent: Path | None = None
    remote_workdir: Path = Path("/tmp")
    local_crc: int = 0
    remote_crc: int = 0
    validated_remote_crc: int = 0
    keeper_context: bool = True  # Eventually share these objects
    remote_object_exists: bool = False
    in_sync: bool = False
    concrete_fs: ConcreteFsEntry | None = None
    # root_name: str = ""

    @property
    def local_path(self):
        return self.local_parent / self.name

    @property
    def empty(self):
        return self.concrete_fs.empty

    @property
    def local_object_exists(self):
        return self.concrete_fs.exists

    @property
    def storable(self):
        return self.concrete_fs.storable

    @property
    def readable(self):
        return self.concrete_fs.readable

    @property
    def absolute_remote_path(self) -> Path:
        # this is ugly, fix
        # if self.concrete_fs.parent:
        #     print("CONCRETE FS PARENT", self.concrete_fs.parent.path)
        # else:
        #     print("NO CONCRETE FS PARENT")
        # print("NAME", self.name)
        # # this is the fileserver root
        # if self.concrete_fs.parent == None:

        #     return self.remote_workdir

        # fileserver
        # if self.root_name:
        #     return self.remote_workdir / self.root_name

        if self.concrete_fs.parent and self.concrete_fs.parent.path.name == "fake_root":
            return "/" / Path(self.name)

        match self.remit.remote_dir:
            case x if x and x.is_absolute():
                return x / self.name
            case x if x:
                return self.remote_workdir / self.remit.remote_dir / self.name
            case x if not x:
                return self.remote_workdir / self.name

    @property
    def absolute_remote_dir(self) -> Path:
        match self.remit.remote_dir:
            case x if x and x.is_absolute():
                return x
            case x if x:
                return self.remote_workdir / self.remit.remote_dir
            case x if not x:
                return self.remote_workdir

    def root(self) -> Path:
        return self.concrete_fs.root().path

    # just use property
    def set_syncronized(self):
        self.in_sync = True

    # BUILD THIS
    def serialize(self):
        ...

    def validate_local_object(self):
        # this doesn't make sense anymore, validation is done as part of instantiation
        # If we're the root object, fake the name for crc
        # if self.concrete_fs.depth == 0:
        #     self.local_crc = self.concrete_fs.crc32(name=self.name)
        # else:
        self.local_crc = self.concrete_fs.crc32()

    def compare_objects(self):
        """"""
        if not self.local_object_exists:
            return

        if not self.remote_crc:  # remote file crc is 0 if it doesn't exist
            self.remote_object_exists = False
            self.in_sync = False

            if self.local_crc:
                log.info(f"Agent needs new object {self.name}... sending")
                return

        self.remote_object_exists = True

        if self.remote_crc != self.local_crc:
            self.in_sync = False
            if (
                self.validated_remote_crc == self.remote_crc
                and self.local_object_exists
            ):
                log.info(
                    f"Agent remote object {self.name} is different than local object but has been validated due to sync strategy"
                )
                return

            if self.local_object_exists:
                # log.info(
                #     f"Remote object {self.name} CRC: {self.remote_crc} Local object CRC: {self.local_crc}... validating difference"
                # )
                return

        if self.remote_crc == self.local_crc:
            # generates too many log messages
            # log.info(f"Agent object {self.name} is up to date... skipping!")
            self.in_sync = True


@dataclass
class FsStateManager:
    """File System State Manager"""

    managed_objects: list[FsEntryStateManager] = field(default_factory=list)
    directives: list[RemoteStateDirective] = field(default_factory=list)
    # local_workdir: Path | None = None
    # remote_workdir: Path | None = None

    def __iter__(self):
        yield from self.managed_objects

    @classmethod
    def filter_hierarchy(
        cls, current_path: Path, existing_paths: list[Path]
    ) -> list[Path]:
        # this needs heavy testing

        paths = existing_paths
        for existing_path in existing_paths:
            if current_path.is_relative_to(existing_path):
                # our ancestor is already in the list. We will get replaced
                # when they get synced (don't add ourselves)
                return paths

            elif existing_path.is_relative_to(current_path):
                # we are higher up the tree.. remove existing_path and
                # install ourselves
                paths.remove(existing_path)
                paths.append(current_path)
                return paths

        # sibling
        paths.append(current_path)

        return paths

    def set_syncronized(self, targets: list[Path]):
        for target in targets:
            if entry := self.get_object_by_local_path(target):
                entry.set_syncronized()

    def get_directive_by_path(self, path: Path) -> RemoteStateDirective | None:
        return next(
            filter(
                lambda x: x.content_source == path.parent and x.name == path.name,
                self.directives,
            ),
            None,
        )

    def get_directive_by_remote_path(self, path: Path) -> RemoteStateDirective:
        return next(
            filter(
                lambda x: x.remote_dir == path.parent and x.name == path.name,
                self.directives,
            ),
            None,
        )

    def add_directives(self, directives: list[RemoteStateDirective]):
        for directive in directives:
            self.add_directive(directive)

    def add_directive(self, directive: RemoteStateDirective):
        if isinstance(directive, RemoteStateDirective):
            self.directives.append(directive)
        else:
            log.warn(
                f"Directive: {directive} is not of type `RemoteStateDirective`. Skipping..."
            )

    def add_objects(self, objects: list):
        for obj in objects:
            self.add_object(obj)

    def add_object(self, fs_object: FsEntryStateManager):
        if isinstance(fs_object, FsEntryStateManager):
            self.managed_objects.append(fs_object)
        else:
            log.warn(
                f"FsObject: {fs_object} is not of type `FsEntryStateManager`. Skipping..."
            )

    def absolute_remote_dirs(self) -> list[Path]:
        return [x.absolute_remote_path for x in self.managed_objects if x.storable]

    def absolute_remote_paths(self) -> list[str]:
        return [str(x.absolute_remote_path) for x in self.managed_objects]

    # def merge_config(self, objects):
    #     for obj in objects:
    #         if isinstance(obj, RemoteStateDirective):
    #             self.managed_objects.append(obj)
    #         else:
    #             log.error(
    #                 f"Object of type {type(obj)} added to managed_objects, must be `RemoteStateDirective`"
    #             )

    def get_object_by_remote_path(self, remote: Path) -> FsEntryStateManager:
        # just lambda
        for fs_object in self.managed_objects:
            path = fs_object.absolute_remote_path

            if path == remote:
                return fs_object

    def get_object_by_local_path(self, local: Path) -> FsEntryStateManager:
        # just lambda
        for fs_object in self.managed_objects:
            path = fs_object.concrete_fs.path

            if path == local:
                return fs_object

    def get_all_objects(self) -> list[FsEntryStateManager]:
        return self.managed_objects

    def update_paths(self, local: Path, remote: Path | None = None):
        self.working_dir = local
        self.remote_workdir = remote

    def validate_local_objects(self):
        for obj in self.managed_objects:
            obj.validate_local_object()


# example (needs updated for parent, fh, last_modified, depth)

# blah = ConcreteFsEntry(
#     Path("/tmp/rangi"),
#     fs_type=FsType.DIRECTORY,
#     size=0,
#     children=[
#         ConcreteFsEntry(
#             Path("/tmp/rangi/bluht.txt"),
#             depth=1,
#             fs_type=FsType.FILE,
#             size=0,
#         ),
#         ConcreteFsEntry(
#             Path("/tmp/rangi/weiner"),
#             depth=1,
#             fs_type=FsType.FILE,
#             size=0,
#         ),
#         ConcreteFsEntry(
#             Path("/tmp/rangi/wrongo"),
#             depth=1,
#             fs_type=FsType.DIRECTORY,
#             size=0,
#             children=[
#                 ConcreteFsEntry(
#                     Path("/tmp/rangi/wrongo.job.exe"),
#                     depth=2,
#                     fs_type=FsType.FILE,
#                     size=0,
#                 )
#             ],
#         ),
#     ],
# )

if __name__ == "__main__":
    blimp = ConcreteFsEntry.build_tree(
        # Path("/Users/davew/.vault/gravyboat/components/fluxagent/fake_root/racing")
        # Path("/Users/davew/.vault")
        Path("/Users/davew/code/flux/fluxvault/LICENSE")
    )
    blimp.realize()
    print(blimp)

    # async def main():
    #     chug = ConcreteFsEntry(
    #         Path("/tmp/rangi/ubu/ubuntu-22.04.1-live-server-amd64.iso"),
    #         depth=0,
    #         fs_type=FsType.UNKNOWN,
    #         size=0,
    #     )
    #     print(await chug.read())
    #     print(await chug.read(5000))
    #     await chug.close()

    # asyncio.run(main())



from dataclasses import dataclass, field
from pathlib import Path

import yaml
from rich.pretty import pprint

from fluxvault.fs import ConcreteFsEntry, FsEntryStateManager, FsStateManager
from fluxvault.helpers import AppMode, RemoteStateDirective, SyncStrategy
from fluxvault.log import log
from fluxvault.extensions import FluxVaultExtensions


@dataclass
class FluxTask:
    name: str
    params: list


## Flow
# Collection of RemoteStateDirective created in CLI. Covers both files
# and folders. Controls how files in physical dirs are synced
# Fs created from fake_root
# loop through local files, match to a statedirective.
# If found, follow that, create RemoteFsEntry and create a ConcreteFs entry and slot it in the tree,
# otherwise slot in tree at local_parent (workdir)
# state directives are attached to RemoteFsEntry.
# EVERY FILE IN FAKE ROOT TURNS INTO A RemoteFsEntry
# loop through unfound stateDirectives and match in fake_root. Update RemoteFsEntry
# Once state has been matched, realize? Then validate?


@dataclass
class FluxComponent:
    name: str
    state_manager: FsStateManager = field(default_factory=FsStateManager)
    tasks: list[FluxTask] | None = None
    # root_dir: Path | None = None
    local_workdir: Path | None = None
    remote_workdir: Path = Path("/tmp")
    member_of: set[str] = field(default_factory=set)

    def add_groups(self, groups: list[str]):
        self.member_of = self.member_of.union(set(groups))

    def update_paths(self, dir: Path):
        # mixing in remote path here
        self.local_workdir = dir
        self.state_manager.update_paths(self.local_workdir, self.remote_workdir)

    def validate_local_objects(self):
        self.state_manager.validate_local_objects()

    def add_directives(self, directives: list[RemoteStateDirective]):
        self.state_manager.add_directives(directives)

    def add_tasks(self, tasks: list):
        for task in tasks:
            self.add_task(task)

    def add_task(self, task: FluxTask):
        self.tasks.append(task)

    def get_task(self, name) -> FluxTask:
        for task in self.tasks:
            if task.name == name:
                return task

    def remove_catalogue(self):
        state_file: Path = self.local_workdir / ".fake_root_state"

        if state_file.is_file():
            with open(state_file, "r") as stream:
                state: dict = yaml.safe_load(stream.read())

            for link in state.get("symlinks", []):
                log.info(f"Unlinking: {link}")
                Path(link).unlink()

            state_file.unlink()

            for dir in state.get("dirs", [])[::-1]:
                # LIFO
                log.info(f"Removing dir: {dir}")
                Path(dir).rmdir()

    def symlink_group(self, group, groups_dir, app_root, fake_root) -> tuple:
        created_dirs = []
        created_symlinks = []

        group_dir = groups_dir / group
        group_dir.mkdir(parents=True, exist_ok=True)

        group_entries = ConcreteFsEntry.entries_in_dir(group_dir)

        for path in group_entries:
            remote_path: Path = self.remote_workdir / path.name

            directive = self.state_manager.get_directive_by_path(
                path.relative_to(app_root)
            )

            if directive:
                if directive.remote_dir.is_absolute():
                    remote_path = directive.remote_dir / path.name
                elif directive.remote_dir:
                    remote_path = self.remote_workdir / directive.remote_dir / path.name

            remote_relative = remote_path.relative_to("/")
            parent_parts = remote_relative.parent.parts

            previous = None
            for dir in parent_parts:
                if not previous:
                    next_dir = fake_root / dir
                else:
                    next_dir = previous / dir
                if not next_dir.is_dir():
                    # will throw if file exists with same name
                    next_dir.mkdir()
                    created_dirs.append(str(next_dir))
                previous = next_dir

            fake_path = fake_root / remote_relative

            if not fake_path.exists():
                fake_path.symlink_to(path)
                created_symlinks.append(str(fake_path))
        return created_dirs, created_symlinks

    def symlink_path(self, path: Path, app_root, fake_root) -> tuple:
        created_dirs = []
        created_symlinks = []

        # default
        remote_path: Path = self.remote_workdir / path.name

        directive = self.state_manager.get_directive_by_path(path.relative_to(app_root))

        if directive:
            if directive.remote_dir.is_absolute():
                remote_path = directive.remote_dir / path.name
            elif directive.remote_dir:
                remote_path = self.remote_workdir / directive.remote_dir / path.name

        remote_relative = remote_path.relative_to("/")
        parent_parts = remote_relative.parent.parts

        previous = None
        for dir in parent_parts:
            if not previous:
                next_dir = fake_root / dir
            else:
                next_dir = previous / dir
            if not next_dir.is_dir():
                # will throw if file exists with same name
                next_dir.mkdir()
                created_dirs.append(str(next_dir))
            previous = next_dir

        fake_path = fake_root / remote_relative
        # fake_path.parent.mkdir(parents=True, exist_ok=True)

        if not fake_path.exists():
            fake_path.symlink_to(path)
            created_symlinks.append(str(fake_path))

        return created_dirs, created_symlinks

    def build_fs(
        self,
        app_mode: AppMode,
        fileserver_dir: Path,
        tree_root: Path,
        remote_prefix: Path = Path("/"),
    ):
        root_tree = ConcreteFsEntry.build_tree(tree_root)
        root_tree.realize()

        log.info("File tree:\n")
        print(root_tree)
        print()

        for fs_object in root_tree.recurse():
            # if fs_object.path.name == tree_root.name:
            if fs_object.path.name == "fake_root":
                continue

            remote_absolute_path = remote_prefix / fs_object.path.relative_to(tree_root)
            directive = self.state_manager.get_directive_by_remote_path(
                remote_absolute_path
            )

            if not directive:
                # Default directive
                if (
                    app_mode == AppMode.FILESERVER
                    and remote_absolute_path.is_relative_to(self.remote_workdir)
                ):
                    sync_strategy = SyncStrategy.STRICT
                else:
                    sync_strategy = SyncStrategy.ENSURE_CREATED

                directive = RemoteStateDirective(
                    remote_dir=remote_absolute_path.parent, sync_strategy=sync_strategy
                )

            # this should only happen with fileserver root
            # root_name = ""
            # if fs_object.path.name == tree_root.name:
            #     root_name = self.remote_workdir.name

            managed_object = FsEntryStateManager(
                name=fs_object.path.name,
                remit=directive,
                local_parent=fs_object.path.parent,
                remote_workdir=self.remote_workdir,
                concrete_fs=fs_object,
                # root_name=root_name,
            )
            self.state_manager.add_object(managed_object)

    def build_catalogue(self, app_mode: AppMode, fileserver_dir: Path, app_root: Path):
        """Builds catalogue for specific component, called by parent `FluxApp`"""

        fake_root: Path = self.local_workdir / "fake_root"
        staging_dir: Path = self.local_workdir / "staging"
        groups_dir: Path = app_root / "groups"

        fake_root.mkdir(parents=True, exist_ok=True)
        staging_dir.mkdir(parents=True, exist_ok=True)
        groups_dir.mkdir(parents=True, exist_ok=True)

        files_in_root = ConcreteFsEntry.are_files_in_dir(fake_root)

        if files_in_root:
            raise ValueError(
                "Files at top level not allowed in fake_root, use a directory (remember to check for hidden files"
            )

        state_file: Path = self.local_workdir / ".fake_root_state"
        # we crashed, or system did etc
        if state_file.is_file():
            self.remove_catalogue()

        created_dirs = []
        created_symlinks = []

        if app_mode == AppMode.FILESERVER:
            entries = ConcreteFsEntry.entries_in_dir(fileserver_dir)
            for entry in entries:
                fake_path = staging_dir / entry.name
                fake_path.symlink_to(entry)
                created_symlinks.append(str(fake_path))
            # leave clean state
            app_dirs = [
                app_root,
                self.local_workdir.parent,
                self.local_workdir,
                groups_dir,
                groups_dir / "all",
                fake_root,
                staging_dir,
            ]
            app_dirs = [str(x) for x in app_dirs]
            created_dirs.extend(app_dirs)

        ConcreteFsEntry.log_objects_in_dir(fake_root)

        # if app_mode == AppMode.FILESERVER:
        # symlink in vault_dir to staging_dir
        # dirs, sims = self.symlink_path(app, app_root, fake_root)

        entries = ConcreteFsEntry.entries_in_dir(staging_dir)

        for group in self.member_of:
            dirs, sims = self.symlink_group(group, groups_dir, app_root, fake_root)
            created_dirs.extend(dirs)
            created_symlinks.extend(sims)

        # do groups first so component overwrites if conflict? (specificity)
        # doesn't seem very efficient
        for path in entries:
            dirs, sims = self.symlink_path(path, app_root, fake_root)
            created_dirs.extend(dirs)
            created_symlinks.extend(sims)

        self.build_fs(app_mode, fileserver_dir, fake_root)

        with open(self.local_workdir / ".fake_root_state", "w") as stream:
            stream.write(
                yaml.dump({"dirs": created_dirs, "symlinks": created_symlinks})
            )


@dataclass
class FluxApp:
    name: str
    components: list[FluxComponent] = field(default_factory=list)
    comms_port: int = 8888
    sign_connections: bool = False
    signing_key: str = ""
    root_dir: Path = field(default_factory=Path)
    fileserver_dir: Path = field(default_factory=Path)
    fluxnode_ips: list[str] = field(default_factory=list)
    app_mode: AppMode = AppMode.MULTI_COMPONENT
    # don't really like that this is here... Crosses a functional line?
    extensions: FluxVaultExtensions = field(default_factory=FluxVaultExtensions)

    def add_component(self, component: FluxComponent):
        existing = next(
            filter(lambda x: x.name == component.name, self.components), None
        )
        if existing:
            log.warn(f"Component already exists: {component.name}")
            return

        component.root_dir = self.root_dir / component.name
        self.components.append(component)

    def ensure_included(self, name: str) -> FluxComponent:
        component = next(filter(lambda x: x.name == name, self.components), None)
        if not component:
            component = FluxComponent(name)
            self.add_component(component)

        return component

    def get_component(self, name: str = "") -> FluxComponent:
        if self.app_mode in [AppMode.FILESERVER, AppMode.SINGLE_COMPONENT]:
            return self.components[0]

        return next(filter(lambda x: x.name == name, self.components), None)

    def merge_global_into_component(self, component: FluxComponent):
        global_config = self.state_manager.get_all_objects()
        component.state_manager.merge_config(global_config)

    def ensure_removed(self, name: str):
        self.components = [c for c in self.components if c.get("name") != name]

    def update_common_objects(self, files: list[FsEntryStateManager]):
        self.state_manager.add_objects(files)

    def update_paths(self, root_app_dir: Path):
        for component in self.components:
            component.update_paths(root_app_dir / "components" / component.name)
        self.state_manager.update_paths(root_app_dir / "common_files")

    def validate_local_objects(self):
        for component in self.components:
            component.validate_local_objects()

    def serialize(self):
        ...

    def remove_catalogue(self):
        """Deletes all symlinks and dirs created in fake_root"""
        for component in self.components:
            component.remove_catalogue()

    def build_catalogue(self):
        for component in self.components:
            component.build_catalogue(
                self.app_mode, self.fileserver_dir.resolve(), self.root_dir
            )

    # def build_fs(self):
    #     # should only ever be one here? This is for FILESERVER ONLY
    #     for component in self.components:
    #         component.build_fs(self.app_mode, self.fileserver_dir, self.root_dir, component.remote_workdir)
